#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>
#include <sys/timeb.h>
#include "interp.h"

// Fixed encode, decode, notl and com
// Fixed coginit
// Implemented reboot

#define OS_COMMAND 0x7ff8
#define OS_PARM    0x7ffc

extern int printflag;
extern CogVarsT CogVars[8];
extern char hubram[0x10000];

extern char lockstate[8];
extern char lockalloc[8];

int ExecuteMathOp(CogVarsT *cogvars, int opcode, int parm1);
int ExecuteExtraOp(CogVarsT *cogvars, int opcode, int parm1, int mask);
void ExecuteRegisterOp(CogVarsT *cogvars, int operand, int msb, int lsb);
void reboot(void);

long GetCnt()
{
    struct timeb timebuf;
    long long millisec;
    long cycles;

    ftime(&timebuf);

    millisec = timebuf.time;
    millisec = (millisec * 1000) + timebuf.millitm;
    millisec *= LONG(0)/1000;
    cycles = millisec;

    return cycles;
}

int GetSignedOffset(long *ppcurr)
{
    int val;
    long pcurr = *ppcurr;
    val = BYTE(pcurr++);
    if (val < 0x80)
    {
	val = (val << 25) >> 25;
    }
    else
    {
	val = (val << 8) | BYTE(pcurr++);
	val = (val << 17) >> 17;
    }
    *ppcurr = pcurr;
    return val;
}

int GetUnsignedOffset(long *ppcurr)
{
    long pcurr = *ppcurr;
    int val;
    val = BYTE(pcurr++);
    if (val & 0x80)
	val = ((val & 0x7f) << 8) | BYTE(pcurr++);
    *ppcurr = pcurr;
    return val;
}

int bitrev(int val, int nbits)
{
    int retval = 0;

    while (nbits-- > 0)
    {
        retval = (retval << 1) | (val & 1);
        val >>= 1;
    }

    return retval;
}

void StartCog(CogVarsT *cogvars, int par, int cogid)
{
    memset(cogvars, 0, sizeof(CogVarsT));
    cogvars->id = cogid;
    cogvars->par = par;
    cogvars->state = 1;
    cogvars->pbase = WORD(par + 2);
    cogvars->vbase = WORD(par + 4);
    cogvars->dbase = WORD(par + 6);
    cogvars->pcurr = WORD(par + 8);
    cogvars->dcurr = WORD(par + 10);
    cogvars->masklong = 0xffffffff;
    cogvars->masktop  = 0x80000000;
    cogvars->maskwr   = 0x00800000;
}

void ExecuteLowerOp(CogVarsT *cogvars)
{
    long pcurr = cogvars->pcurr;
    long dcurr = cogvars->dcurr;
    int opcode = BYTE(pcurr++);

    if (opcode <= 3) // ldfrmr, ldfrm, ldfrmar, ldfrma
    {
	opcode |= cogvars->pbase;
	WORD(dcurr) = opcode;
	dcurr += 2;
	WORD(dcurr) = cogvars->vbase;
	dcurr += 2;
	WORD(dcurr) = cogvars->dbase;
	dcurr += 2;
	WORD(dcurr) = cogvars->dcall;
	cogvars->dcall = dcurr;
	dcurr += 2;
	LONG(dcurr) = 0;
	dcurr += 4;
    }
    else if (opcode == 0x04) // jmp
    {
	int jmpoff = GetSignedOffset(&pcurr);
	pcurr += jmpoff;
    }
    else if (opcode >= 0x05 && opcode <= 0x07) // call, callobj, callobjx
    {
	int index, pubnum;

        if (opcode > 0x05)
	{
	    int objnum = BYTE(pcurr++);
	    if (opcode == 0x07)
	    {
		dcurr -= 4;
		objnum += LONG(dcurr);
	    }
	    index = (objnum << 2) + cogvars->pbase;
	    cogvars->pbase += WORD(index);
	    cogvars->vbase += WORD(index+2);
	}

	pubnum = BYTE(pcurr++);
        cogvars->dbase = cogvars->dcall;
	cogvars->dcall = WORD(cogvars->dcall);
	WORD(cogvars->dbase) = pcurr;
	index = (pubnum << 2) + cogvars->pbase;
	pcurr = WORD(index) + cogvars->pbase;
	cogvars->dbase += 2;
	dcurr += WORD(index+2);
    }
    else if (opcode == 0x08) // tjz
    {
	int jmpoff = GetSignedOffset(&pcurr);

	if (!LONG(dcurr-4))
	{
	    pcurr += jmpoff;
	    dcurr -= 4;
	}
    }
    else if (opcode == 0x09) // djnz
    {
	int jmpoff = GetSignedOffset(&pcurr);
	//printf("djnz: %d\n", LONG(dcurr-4));
	if (--LONG(dcurr-4))
	{
	    pcurr += jmpoff;
	}
	else
	    dcurr -= 4;
    }
    else if (opcode == 0x0a) // jz
    {
	int jmpoff = GetSignedOffset(&pcurr);
	dcurr -= 4;
	if (!LONG(dcurr))
	{
	    pcurr += jmpoff;
	}
    }
    else if (opcode == 0x0b) // jnz
    {
	int jmpoff = GetSignedOffset(&pcurr);
	dcurr -= 4;
	if (LONG(dcurr))
	{
	    pcurr += jmpoff;
	}
    }
    else if (opcode >= 0x0c && opcode <= 0x15)
    {
	int val, jmpoff, val2, val0;
	if (opcode == 0x0c) // casedone
	{
	    dcurr -= 8;
	    jmpoff = LONG(dcurr);
	    pcurr = cogvars->pbase + jmpoff;
	}
	else if (opcode == 0x0d) // casevalue
	{
	    jmpoff = GetSignedOffset(&pcurr);
	    dcurr -= 4;
	    val = LONG(dcurr);
	    if (val == LONG(dcurr - 4))
	    {
		pcurr += jmpoff;
	    }
	}
	else if (opcode == 0x0e) // caserange
	{
	    jmpoff = GetSignedOffset(&pcurr);
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    val2 = LONG(dcurr);
	    val0 = LONG(dcurr - 4);
	    if (val <= val2)
	    {
		if (val <= val0 && val0 <= val2) pcurr += jmpoff;
	    }
	    else
	    {
		if (val2 <= val0 && val0 <= val) pcurr += jmpoff;
	    }
	}
	else if (opcode == 0x0f) // lookdone
	{
	    dcurr -= 8;
	    LONG(dcurr - 4) = 0;
	}
	else if (opcode == 0x10) // lookupval
	{
	    dcurr -= 4;
	    val = LONG(dcurr);
	    val0 = LONG(dcurr - 4);
	    if (val0 < LONG(dcurr - 12))
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
		LONG(dcurr - 4) = 0;
	    }
	    else if (val0 == LONG(dcurr - 12))
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
		LONG(dcurr - 4) = val;
	    }
	    else
	    {
		LONG(dcurr - 12)++;
	    }
	}
	else if (opcode == 0x11) // lookdnval
	{
	    dcurr -= 4;
	    val = LONG(dcurr);
	    val0 = LONG(dcurr - 4);
	    if (val == val0)
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
	    }
	    else
	    {
		LONG(dcurr - 12)++;
	    }
	}
	else if (opcode == 0x12) // lookuprng
	{
	    int num, count;
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    val2 = LONG(dcurr);
	    val0 = LONG(dcurr - 4);
	    num = val - val2;
	    if (num < 0) num = -num;
	    count = LONG(dcurr - 12);
	    //printf("\nlookuprng: %d %d %d %d %d\n", val0, count, val2, val, num);
	    if (val0 < count)
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
		LONG(dcurr - 4) = 0;
	    }
	    else if (val0 - num <= count)
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
		num = val0 - count;
		//printf("\nnum = %d\n", num);
		if (val >= val2)
		{
		    LONG(dcurr - 4) = val2 + num;
		}
		else
		{
		    LONG(dcurr - 4) = val - num;
		}
	    }
	    else
	    {
		LONG(dcurr - 12) += num + 1;
	    }
	}
	else if (opcode == 0x13) // lookdnrng
	{
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    val2 = LONG(dcurr);
	    if (val > val2)
	    {
		val0 = val;
		val = val2;
		val2 = val0;
	    }
	    val0 = LONG(dcurr - 4);
	    if (val <= val0 && val0 <= val2)
	    {
		dcurr -= 8;
		pcurr = cogvars->pbase + LONG(dcurr);
	    }
	    else
	    {
		LONG(dcurr - 12)++;
	    }
	}
	else if (opcode == 0x14) // pop
	{
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= val;
	}
	else if (opcode == 0x15) // run
	{
	    cogvars->lsb = pcurr;
	    pcurr = 0xfffc;
	}
	else
	    printf("%4.4x %2.2x - NOT IMPLEMENTED\n", pcurr - 1, opcode);
    }
    else if (opcode >= 0x16 && opcode <= 0x23)
    {
        if (opcode == 0x16) // strsize
        {
	    char *ptr = hubram;
	    dcurr -= 4;
	    LONG(dcurr) = strlen(ptr + LONG(dcurr));
	    dcurr += 4;
        }
        else if (opcode == 0x17) // strcomp
        {
	    char *ptr1 = hubram;
	    char *ptr2 = hubram;
	    dcurr -= 4;
	    ptr1 += LONG(dcurr);
	    dcurr -= 4;
	    ptr2 += LONG(dcurr);
	    if (strcmp(ptr1, ptr2) == 0)
	        LONG(dcurr) = -1;
	    else
	        LONG(dcurr) = 0;
	    dcurr += 4;
        }
        else if (opcode == 0x18) // bytefill
        {
	    int val, num;
	    char *ptr = hubram;
	    dcurr -= 4;
	    num = LONG(dcurr);
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    ptr += LONG(dcurr);
	    memset(ptr, val, num);
        }
        else if (opcode == 0x19) // wordfill
        {
	    int val, num, addr;
	    dcurr -= 4;
	    num = LONG(dcurr);
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    addr = LONG(dcurr);
	    while (num-- > 0)
	    {
		WORD(addr) = val;
		addr += 2;
	    }
        }
        else if (opcode == 0x1a) // longfill
        {
	    int val, num, addr;
	    dcurr -= 4;
	    num = LONG(dcurr);
	    dcurr -= 4;
	    val = LONG(dcurr);
	    dcurr -= 4;
	    addr = LONG(dcurr);
	    while (num-- > 0)
	    {
		LONG(addr) = val;
		addr += 4;
	    }
        }
        else if (opcode == 0x1b) // waitpeq
        {
	    dcurr -= 12;
        }
        else if (opcode >= 0x1c && opcode <= 0x1e ) // bytemove, wordmove, longmove
        {
	    char *dst = hubram;
	    char *src = hubram;
	    int num;
	    dcurr -= 4;
	    num = LONG(dcurr);
	    dcurr -= 4;
	    src += LONG(dcurr);
	    dcurr -= 4;
	    dst += LONG(dcurr);
	    if (opcode == 0x1d)
	        num <<= 1;
	    else if (opcode == 0x1e)
	        num <<= 2;
	    memmove(dst, src, num);
        }
        else if (opcode == 0x1f) // waitpne
        {
	    dcurr -= 12;
        }
        else if (opcode == 0x20) // clkset
        {
	    dcurr -= 4;
	    LONG(0) = LONG(dcurr);
	    dcurr -= 4;
	    BYTE(4) = LONG(dcurr);
        }
        else if (opcode == 0x21) // cogstop
        {
	    int cogid;
	    dcurr -= 4;
	    cogid = LONG(dcurr) & 7;
	    CogVars[cogid].state = 0;
	    if (LONG(0) == 0 && BYTE(4) == 0x80)
	    {
		reboot();
		return;
	    }
	    //printf("COG %d: cogstop %d\n", cogvars->id, cogid);
        }
        else if (opcode == 0x22) // lockret
        {
	    int locknum;

	    dcurr -= 4;
	    locknum = LONG(dcurr) & 7;
	    lockalloc[locknum] = 0;
        }
        else if (opcode == 0x23) // waitcnt
        {
	    int parm1;

	    dcurr -= 4;
	    parm1 = LONG(dcurr) - GetCnt();
	    if (parm1 < 0 || parm1 > 200000)
	    {
		pcurr--;
		dcurr += 4;
		cogvars->state = 2;
	    }
	    else
	        cogvars->state = 1;
        }
    }
    else if (opcode >= 0x24 && opcode <= 0x2f)
    {
	if (opcode >= 0x24 && opcode <= 0x26) // ldregx, stregx, exregx
	{
	    int operand;
	    dcurr -= 4;
	    operand = LONG(dcurr);
	    operand = ((opcode & 3) << 5) | (operand & 15);
            cogvars->pcurr = pcurr;
            cogvars->dcurr = dcurr;
	    ExecuteRegisterOp(cogvars, operand, 31, 0);
            pcurr = cogvars->pcurr;
            dcurr = cogvars->dcurr;
	}
	else if (opcode == 0x27) // waitvid
	{
	    dcurr -= 8;
	}
	else if (opcode == 0x28 || opcode == 0x2c) // coginitret, coginit
	{
	    int cogid, addr, par;
	    dcurr -= 4;
	    par = LONG(dcurr);
	    dcurr -= 4;
	    addr = LONG(dcurr);
	    dcurr -= 4;
	    cogid = LONG(dcurr);
	    if (cogid < 0 || cogid > 7)
	    {
		for (cogid = 0; cogid < 8; cogid++)
		{
		    if (!CogVars[cogid].state) break;
		}
	    }
	    if (cogid < 0 || cogid > 7 || addr != 0xf004)
	        cogid = -1;
	    if (opcode == 0x28)
	    {
	        //printf("coginitret(%d, %4.4x, %4.4x)\n", cogid, addr, par);
	        LONG(dcurr) = cogid;
	        dcurr += 4;
	    }
	    //else
	        //printf("coginit(%d, %4.4x, %4.4x)\n", cogid, addr, par);
	    if (cogid != -1)
	    {
                cogvars->pcurr = pcurr;
                cogvars->dcurr = dcurr;
	        StartCog(&CogVars[cogid], par, cogid);
                pcurr = cogvars->pcurr;
                dcurr = cogvars->dcurr;
	    }
	}
	else if (opcode == 0x29 || opcode == 0x2d) // locknewret, locknew
	{
	    int locknum;

	    for (locknum = 0; locknum < 8; locknum++)
	    {
		if (!lockalloc[locknum])
		{
		    lockalloc[locknum] = 1;
		    break;
		}
	    }
	    if (opcode == 0x29)
	    {
		if (locknum == 8) locknum = -1;
		LONG(dcurr) = locknum;
		dcurr += 4;
	    }
	}
	// locksetret, lockclrret, lockset, lockclr
	else if (opcode == 0x2a || opcode == 0x2b || opcode == 0x2e || opcode == 0x2f)
	{
	    int locknum;

	    dcurr -= 4;
	    locknum = LONG(dcurr) & 7;
	    if (opcode <= 0x2b)
	    {
	        LONG(dcurr) = lockstate[locknum];
	        dcurr += 4;
	    }
	    lockstate[locknum] = (opcode & 1) - 1;
	}
    }
    else if (opcode >= 0x30 && opcode <= 0x33) // abort, abortval, ret, retval
    {
	int retval, dbase, pbase;

	if (opcode & 1)
	    retval = LONG(dcurr - 4);
	else
	    retval = LONG(cogvars->dbase);

        dbase = cogvars->dbase;

        while (1)
	{
	    dcurr = dbase - 8;
	    pbase = WORD(dcurr);
	    dbase = WORD(dcurr + 4);
	    if ((opcode & 2) || (pbase & 2)) break;
	}

	cogvars->pbase = pbase & 0xfffc;
	cogvars->vbase = WORD(dcurr + 2);
	cogvars->dbase = dbase;
	pcurr = WORD(dcurr + 6);

        if (!(pbase & 1))
	{
	    LONG(dcurr) = retval;
	    dcurr += 4;
	}
    }
    else if (opcode >= 0x34 && opcode < 0x3c)
    {
	if (opcode == 0x35) LONG(dcurr) = 0;        // dli0
	else if (opcode == 0x36) LONG(dcurr) = 1;   // dli1
	else if (opcode == 0x34) LONG(dcurr) = -1;  // dlim1
	else if (opcode == 0x37) // ldlip
	{
	    int operand = BYTE(pcurr++);
	    int val;

	    val = 2 << (operand & 31);
	    if (operand & 0x20) val--;
	    if (operand & 0x40) val = ~val;
	    LONG(dcurr) = val;
	}
	else // ldbi, ldwi, ldmi, ldli
	{
	    int operand = 0;
	    while (opcode-- >= 0x38) operand = (operand << 8) | BYTE(pcurr++);
	    LONG(dcurr) = operand;
	}
	dcurr += 4;
    }
    else if (opcode == 0x3d) // ldregbit, stregbit, exregbit
    {
	int operand = BYTE(pcurr++);
	int bitpos;

	dcurr -= 4;
	bitpos = LONG(dcurr) & 0x1f;

        cogvars->pcurr = pcurr;
        cogvars->dcurr = dcurr;
	ExecuteRegisterOp(cogvars, operand, bitpos, bitpos);
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else if (opcode == 0x3e) // ldregbits, stregbits, exregbits
    {
	int operand = BYTE(pcurr++);
	int msb, lsb;

	dcurr -= 4;
	lsb = LONG(dcurr) & 0x1f;
	dcurr -= 4;
	msb = LONG(dcurr) & 0x1f;

        cogvars->pcurr = pcurr;
        cogvars->dcurr = dcurr;
	ExecuteRegisterOp(cogvars, operand, msb, lsb);
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else if (opcode == 0x3f) // ldreg, streg, exreg
    {
	int operand = BYTE(pcurr++);

        cogvars->pcurr = pcurr;
        cogvars->dcurr = dcurr;
	ExecuteRegisterOp(cogvars, operand, 31, 0);
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else
    {
        printf("NOT PROCESSED\n");
    }
    cogvars->pcurr = pcurr;
    cogvars->dcurr = dcurr;
}

void ExecuteRegisterOp(CogVarsT *cogvars, int operand, int msb, int lsb)
{
    int opcode;
    int parm1, parm2;
    int pcurr = cogvars->pcurr;
    int dcurr = cogvars->dcurr;
    long *reg = (long *)cogvars;
    int memfunc = (operand >> 5) & 3;
    int revflag = 0;
    int mask, nbits;

    if (lsb > msb)
    {
	revflag = msb;
	msb = lsb;
	lsb = revflag;
	revflag = 1;
    }

    nbits = msb - lsb + 1;

    if (nbits >= 32)
        mask = -1;
    else
        mask = (1 << nbits) - 1;

    operand &= 0x1f;

    if (memfunc == 1) // store
    {
        cogvars->dcurr -= 4;
	if (nbits == 32 && !revflag)
	{
            reg[operand] = LONG(cogvars->dcurr);
	}
	else
	{
            parm1 = reg[operand];
            parm2 = LONG(cogvars->dcurr);
	    parm2 &= mask;
	    if (revflag) parm2 = bitrev(parm2, nbits);
	    reg[operand] = (parm1 & ~(mask << lsb)) | (parm2 << lsb);
	}
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else if (memfunc == 0) // load
    {
        if (operand == 0x11) // cnt = $1f1
            parm1 = GetCnt();
        else
            parm1 = reg[operand];
	parm1 = (parm1 >> lsb) & mask;
	if (revflag) parm1 = bitrev(parm1, nbits);
	LONG(dcurr) = parm1;
        dcurr += 4;
    }
    else if (memfunc == 2) // execute
    {
        int loadflag;

        opcode = BYTE(cogvars->pcurr++);

        if (opcode & 0x7e)
        {
            if (operand == 0x11) // cnt = $1f1
                parm1 = GetCnt();
            else
                parm1 = reg[operand];
        }

	parm2 = (parm1 >> lsb) & mask;
	if (revflag) parm2 = bitrev(parm2, nbits);

        parm2 = ExecuteExtraOp(cogvars, opcode, parm2, mask);

        parm2 &= mask;
        if (revflag) parm2 = bitrev(parm2, nbits);
        reg[operand] = (parm1 & ~(mask << lsb)) | (parm2 << lsb);

        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else
        printf("Undefined register operation\n");

    cogvars->pcurr = pcurr;
    cogvars->dcurr = dcurr;
}

void ExecuteMemoryOp(CogVarsT *cogvars)
{
    long pcurr = cogvars->pcurr;
    long dcurr = cogvars->dcurr;
    int opcode, memfunc, memsize, membase, memaddr;

    opcode = BYTE(pcurr++);
    memfunc = opcode & 3;

    if (opcode < 0x80) // Compact offset
    {
	memsize = 3;
	memaddr = (opcode & 0x1c);
	membase = (opcode >> 5) & 3;
    }
    else
    {
	memsize = ((opcode >> 5) & 3) + 1;
	membase = (opcode & 0x0c) >> 2;

        // Check for index op
        if (opcode & 0x10)
        {
	    dcurr -= 4;
	    memaddr = LONG(dcurr) << (memsize - 1);
        }
	else
	    memaddr = 0;

	if (membase)
	    memaddr += GetUnsignedOffset(&pcurr);
	else
        {
	    dcurr -= 4;
	    memaddr += LONG(dcurr);
        }
    }

    if (membase == 1)
	memaddr += cogvars->pbase;
    else if (membase == 2)
	memaddr += cogvars->vbase;
    else if (membase == 3)
	memaddr += cogvars->dbase;

    if (memfunc == 3)      // la
    {
	LONG(dcurr) = memaddr;
	dcurr += 4;
    }
    else if (memfunc == 0) // ld
    {
	if (memsize == 1) LONG(dcurr) = BYTE(memaddr);
	else if (memsize == 2) LONG(dcurr) = WORD(memaddr);
	else LONG(dcurr) = LONG(memaddr);
	dcurr += 4;
    }
    else if (memfunc == 1) // st
    {
	dcurr -= 4;
	if (memsize == 1) BYTE(memaddr) = LONG(dcurr);
	else if (memsize == 2) WORD(memaddr) = LONG(dcurr);
	else LONG(memaddr) = LONG(dcurr);
    }
    else                    // ex
    {
	//int loadflag;
        int parm1, parm2;
	opcode = BYTE(pcurr++);
	//loadflag = opcode & 0x80;
	//opcode &= 0x7f;

        if (opcode & 0x7f)
	{
	    if (memsize == 1) parm1 = BYTE(memaddr);
	    else if (memsize == 2) parm1 = WORD(memaddr);
	    else parm1 = LONG(memaddr);
	}

        cogvars->pcurr = pcurr;
        cogvars->dcurr = dcurr;
        parm1 = ExecuteExtraOp(cogvars, opcode, parm1, -1);
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;

	if (memsize == 1) BYTE(memaddr) = parm1;
	else if (memsize == 2) WORD(memaddr) = parm1;
	else LONG(memaddr) = parm1;

	//if (loadflag)
	//{
	    //LONG(dcurr) = parm2;
	    //dcurr += 4;
	//}
    }
    cogvars->pcurr = pcurr;
    cogvars->dcurr = dcurr;
}

int ExecuteExtraOp(CogVarsT *cogvars, int opcode, int parm1, int mask)
{
    int parm2;
    long pcurr = cogvars->pcurr;
    long dcurr = cogvars->dcurr;
    int loadflag = opcode & 0x80;
    int size = (opcode >> 1) & 3;

    if (size == 1)
        mask = 0xff;
    else if (size == 2)
        mask = 0xffff;
    else if (size == 2)
        mask = 0xffffffff;

    opcode &= 0x7f;

    if (opcode >= 0x40 && opcode < 0x60) // math op
    {
        parm2 = parm1 = ExecuteMathOp(cogvars, opcode, parm1);
        pcurr = cogvars->pcurr;
        dcurr = cogvars->dcurr;
    }
    else if ((opcode & 0x7e) == 0x00) // store
    {
        dcurr -= 4;
        parm2 = parm1 = LONG(dcurr);
    }
    else if ((opcode & 0x7a) == 0x02) // repeat, repeats
    {
        int first, last, step, jmpoff;

        dcurr -= 4;
        last = LONG(dcurr);
        dcurr -= 4;
        first = LONG(dcurr);
        if (opcode == 0x06)
        {
	    dcurr -= 4;
	    step = LONG(dcurr);
        }
        else step = 1;
        jmpoff = GetSignedOffset(&pcurr);
        if (last >= first)
        {
	    parm1 += step;
	    if (parm1 >= first && parm1 <= last)
	        pcurr += jmpoff;
        }
        else
        {
	    parm1 -= step;
	    if (parm1 <= first && parm1 >= last)
	        pcurr += jmpoff;
        }
        parm2 = parm1;
    }
    else if ((opcode & 0x78) == 8) // randf, randr
    {
	unsigned int a, c, x, y, z;

        x = parm1;
	z = (opcode & 0x04) == 0;
	if (!x) x = 1;
	y = 32;
	a = 0x17;
	if (!z) a = (a >> 1) | (a << 31);
	while (y--)
	{
	    c = x & a;
	    while (c & 0xfffffffe) c = (c >> 1) ^ (c & 1);
	    if (z)
	        x = (x >> 1) | (c << 31);
	    else
	        x = (x << 1) | c;
	}
        parm2 = parm1 = x;
    }
    else if ((opcode & 0x7c) == 0x10) // sexb
    {
        parm2 = parm1 = (parm1 << 24) >> 24;
    }
    else if ((opcode & 0x7c) == 0x14) // sexw
    {
        parm2 = parm1 = (parm1 << 16) >> 16;
    }
    else if ((opcode & 0x7c) == 0x18) // postclr
    {
	parm2 = parm1;
	parm1 = 0;
    }
    else if ((opcode & 0x7c) == 0x1c) // postset
    {
	parm2 = parm1;
	parm1 = -1;
    }
    else if ((opcode & 0x78) == 0x20) // preinc
    {
        parm2 = ++parm1;
	parm2 &= mask;
    }
    else if ((opcode & 0x78) == 0x28) // postinc
    {
        parm2 = parm1++;
	parm2 &= mask;
    }
    else if ((opcode & 0x78) == 0x30) // predec
    {
        parm2 = --parm1;
	parm2 &= mask;
    }
    else if ((opcode & 0x78) == 0x38) // postdec
    {
        parm2 = parm1--;
	parm2 &= mask;
    }
    else
    {
        printf("NOT IMPLEMENTED\n");
    }

    if (loadflag)
    {
        LONG(dcurr) = parm2;
        dcurr += 4;
    }

    cogvars->pcurr = pcurr;
    cogvars->dcurr = dcurr;

    return parm1;
}

int ExecuteMathOp(CogVarsT *cogvars, int opcode, int parm1)
{
    int parm2, parm3;
    int execflag = 0;
    int unary = 0;
    long dcurr = cogvars->dcurr;

    if (opcode < 0xe0)
    {
	execflag = 1;
	opcode += 0xe0 - 0x40;
    }

    unary = (opcode == 0xe6 || opcode == 0xe7 || opcode == 0xe9 ||
        opcode == 0xf1 || opcode == 0xf3 || opcode == 0xf8 || opcode == 0xff);

    // Get the parameters from the stack
    if (execflag)
    {
	if (!unary)
	{
            dcurr -= 4;
            parm2 = LONG(dcurr);
	}
    }
    else
    {
        if (!unary)
        {
            dcurr -= 4;
            parm2 = LONG(dcurr);
	    dcurr -= 4;
	    parm1 = LONG(dcurr);
        }
        else
	{
            dcurr -= 4;
            parm1 = LONG(dcurr);
	}
    }

    // Execute the math op
    switch (opcode)
    {
	case 0xe0: // ror
        parm1 = (((unsigned long)parm1) >> parm2) | (parm1 << (32 - parm2));
	break;

	case 0xe1: // rol
        parm1 = (((unsigned long)parm1) >> (32 - parm2)) | (parm1 << parm2);
	break;

	case 0xe2: // shr
        parm1 = ((unsigned long)parm1) >> parm2;
	break;

	case 0xe3: // shl
        parm1 <<= parm2;
	break;

	case 0xe4: // min
	if (parm2 > parm1) parm1 = parm2;
	break;

	case 0xe5: // max
	if (parm2 < parm1) parm1 = parm2;
	break;

	case 0xe6: // neg
        parm1 = -parm1;
	break;

	case 0xe7: // com
        parm1 = ~parm1;
	break;

	case 0xe8: // and
        parm1 &= parm2;
	break;

	case 0xe9: // abs
        if (parm1 < 0) parm1 = -parm1;
	break;

	case 0xea: // or
        parm1 |= parm2;
	break;

	case 0xeb: // xor
        parm1 ^= parm2;
	break;

	case 0xec: // add
        parm1 += parm2;
	break;

	case 0xed: // sub
        parm1 -= parm2;
	break;

	case 0xee: // sar
        parm1 >>= parm2;
	break;

	case 0xef: // rev
	parm1 = bitrev(parm1, parm2);
	break;

	case 0xf0: // andl
	parm1 = (parm1 && parm2) ? -1 : 0;
	break;

	case 0xf1: // encode
	for (parm2 = 32; parm2 >= 1; parm2--)
	{
	    if (parm1 & 0x80000000) break;
	    parm1 <<= 1;
	}
	parm1 = parm2;
	break;

	case 0xf4: // mul
        parm1 *= parm2;
	break;

	case 0xf5: // mulh
        {
	    long long parm1a = parm1;
	    long long parm2a = parm2;
	    parm1 = (parm1a * parm2a) >> 32;
        }
	break;

	case 0xf2: // orl
        parm1 = (parm1 || parm2) ? -1 : 0;
	break;

	case 0xf3: // decode
        parm1 = 1 << parm1;
	break;

	case 0xf6: // div
        parm1 /= parm2;
	break;

	case 0xf7: // mod
        parm1 %= parm2;
	break;

	case 0xf8: // sqrt
        parm2 = 0;
        parm3 = 1 << 30;
        while (parm3)
        {
	    parm2 |= parm3;
	    if (parm2 <= parm1)
	    {
	        parm1 -= parm2;
	        parm2 += parm3;
	    }
	    else
	        parm2 -= parm3;
	    parm2 >>= 1;
	    parm3 >>= 2;
        }
	parm1 = parm2;
	break;

	case 0xf9: // cmplt
        parm1 = (parm1 < parm2) ? -1 : 0;
	break;

	case 0xfa: // cmpgt
        parm1 = (parm1 > parm2) ? -1 : 0;
	break;

	case 0xfb: // cmpne
        parm1 = (parm1 != parm2) ? -1 : 0;
	break;

	case 0xfc: // cmpeq
        parm1 = (parm1 == parm2) ? -1 : 0;
	break;

	case 0xfd: // cmple
        parm1 = (parm1 <= parm2) ? -1 : 0;
	break;

	case 0xfe: // cmpgr
        parm1 = (parm1 >= parm2) ? -1 : 0;
	break;

	case 0xff: // notl
        parm1 = parm1 ? 0 : -1;
	break;

	default:
        printf("NOT PROCESSED\n");
    }

    // Push the result back to the stack
    if (!execflag)
    {
        LONG(dcurr) = parm1;
        dcurr += 4;
        cogvars->pcurr++;
    }

    cogvars->dcurr = dcurr;
    return parm1;
}

void ExecuteOp(CogVarsT *cogvars)
{
    int opcode;

    if (!cogvars->state) return;

    opcode = BYTE(cogvars->pcurr);

    if (opcode < 0x40)
        ExecuteLowerOp(cogvars);
    else if (opcode < 0xe0)
        ExecuteMemoryOp(cogvars);
    else
        ExecuteMathOp(cogvars, opcode, 0);
}
