#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>
#include <direct.h>
#include "interp.h"

#define SYS_COMMAND    0xf004
#define SYS_LOCKNUM    0xf006
#define SYS_PARM       0xf008

#define SYS_CON_PUTCH     1
#define SYS_CON_GETCH     2
#define SYS_FILE_OPEN     3
#define SYS_FILE_CLOSE    4
#define SYS_FILE_READ     5
#define SYS_FILE_WRITE    6
#define SYS_FILE_OPENDIR  7
#define SYS_FILE_CLOSEDIR 8
#define SYS_FILE_READDIR  9
#define SYS_FILE_SEEK     10
#define SYS_FILE_TELL     11
#define SYS_FILE_REMOVE   12

char hubram[0x10000];
char lockstate[8];
char lockalloc[8];
long reg[32] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

char objname[100][20];
int methodnum[100];
int methodlev = 0;

int printflag = 0;
int symflag = 0;

void PrintOp(CogVarsT *cogvars);
void ExecuteOp(CogVarsT *cogvars);
int ExecuteMathOp(int opcode, int parm1);
void ReadOpsFile(void);
char *FindChar(char *str, int val);

void usage(void)
{
    printf("usage: spinsim [-l] [-#] file\n");
    exit(1);
}

void putchx(int val)
{
    putchar(val);
    fflush(stdout);
}

int getchx(void)
{
    unsigned char val;
    fread(&val, 1, 1, stdin);
    if (val == 10) val = 13;
    return val;
}

void CheckCommand(void)
{
    int parm;
    int command = WORD(SYS_COMMAND);
    FILE *stream;
    DIR *pdir;
    struct dirent *pdirent;

    if (!command) return;

    parm = LONG(SYS_PARM);

    if (command == SYS_CON_PUTCH)
    {
	if (parm == 13)
	    putchx(10);
	else
	    putchx(parm);
    }
    else if (command == SYS_CON_GETCH)
    {
	if (kbhit())
	    parm = getch();
	else
	    parm = -1;
	LONG(SYS_PARM) = parm;
    }
    else if (command == SYS_FILE_OPEN)
    {
	char *fname;
	char *fmode;

	fname = (char *)&BYTE(LONG(parm));
	fmode = (char *)&BYTE(LONG(parm+4));
	stream = fopen(fname, fmode);
	LONG(SYS_PARM) = (long)stream;
	//printf("OPEN: %s %s %8.8x\n", fname, fmode, stream);
    }
    else if (command == SYS_FILE_CLOSE)
    {
	if (parm) fclose((FILE *)parm);
	LONG(SYS_PARM) = 0;
    }
    else if (command == SYS_FILE_READ)
    {
	int num;
	char *buffer;

         if (!parm) LONG(SYS_PARM) = -1;
	 else
	 {
	    stream = (FILE *)LONG(parm);
	    buffer = (char *)&BYTE(LONG(parm+4));
	    num = LONG(parm+8);
	    //printf("READ: %8.8x %8.8x %d\n", stream, buffer, num);
	    LONG(SYS_PARM) = fread(buffer, 1, num, stream);
	}
    }
    else if (command == SYS_FILE_WRITE)
    {
	int num;
	char *buffer;

	if (!parm) LONG(SYS_PARM) = -1;
	else
	{
	    stream = (FILE *)LONG(parm);
	    buffer = (char *)&BYTE(LONG(parm+4));
	    num = LONG(parm+8);
	    LONG(SYS_PARM) = fwrite(buffer, 1, num, stream);
	}
    }
    else if (command == SYS_FILE_OPENDIR)
    {
	char *dname = ".";

	//dname = (char *)&BYTE(LONG(parm));
	pdir = opendir(dname);
	LONG(SYS_PARM) = (long)pdir;
	//printf("OPENDIR: %8.8x\n", pdir);
    }
    else if (command == SYS_FILE_CLOSEDIR)
    {
	if (parm) closedir((DIR *)parm);
	LONG(SYS_PARM) = 0;
	//printf("CLOSEDIR: %8.8x\n", parm);
    }
    else if (command == SYS_FILE_READDIR)
    {
	int *buffer;
        pdir = (DIR *)LONG(parm);
        buffer = (int *)&BYTE(LONG(parm+4));
	//printf("READDIR: %8.8x\n", pdir);
	if (!pdir) LONG(SYS_PARM) = 0;
	else
	{
	    pdirent = readdir(pdir);
	    if (pdirent)
	    {
	        buffer[0] = pdirent->d_size;
	        buffer[1] = pdirent->d_attr;
	        strcpy((char *)&buffer[2], pdirent->d_name);
	        //printf("fname = %s\n", pdirent->d_name);
	    }
	    LONG(SYS_PARM) = (long)pdirent;
	}
    }
    else if (command == SYS_FILE_SEEK)
    {
	int offset, whence;
        stream = (FILE *)LONG(parm);
        offset = LONG(parm+4);
        whence = LONG(parm+8);
	LONG(SYS_PARM) = fseek(stream, offset, whence);
    }
    else if (command == SYS_FILE_TELL)
    {
        stream = (FILE *)parm;
	LONG(SYS_PARM) = ftell(stream);
    }
    else if (command == SYS_FILE_REMOVE)
    {
	char *fname = (char *)&BYTE(parm);
	LONG(SYS_PARM) = remove(fname);
    }
    WORD(SYS_COMMAND) = 0;
}

void PrintStack(CogVarsT *cogvars)
{
    long dcurr = cogvars->dcurr;
    printf("PrintStack: %4.4x %8.8x %8.8x %8.8x\n",
        dcurr, LONG(dcurr-4), LONG(dcurr-8), LONG(dcurr-12));
}

CogVarsT CogVars[8];

char *bootfile;

void reboot(void)
{
    int i;
    int dbase;
    char *ptr;
    FILE *infile;

    memset(hubram, 0, 32768);
    memset(lockstate, 0, 8);
    memset(lockalloc, 0, 8);

    infile = fopen(bootfile, "rb");

    if (infile == 0)
    {
	printf("Could not open %s\n", bootfile);
	exit(1);
    }

    fread(hubram, 1, 32768, infile);
    fclose(infile);

    // Read the ROM file
    infile = fopen("rom.bin", "rb");

    if (infile == 0)
    {
	printf("Could not open rom.bin\n");
	exit(1);
    }

    fread(hubram + 32768, 1, 32768, infile);
    fclose(infile);

    dbase = WORD(10);
    LONG(dbase-8) = 0xfff9ffff;
    LONG(dbase-4) = 0xfff9ffff;
    LONG(dbase) = 0;

    WORD(SYS_COMMAND) = 0;
    WORD(SYS_LOCKNUM) = 1;
    lockalloc[0] = 1;

    for (i = 0; i < 8; i++) CogVars[i].state = 0;

    StartCog(&CogVars[0], 4, 0);

    strcpy(objname[0], "xxx");
    strcpy(objname[1], bootfile);
    ptr = FindChar(objname[1], '.');
    *ptr = 0;
    strcat(objname[1], ".spn");
    methodnum[0] = 1;
    methodnum[1] = 1;
    methodlev = 1;

    LONG(0xf00c) = printflag;
}

void main(int argc, char **argv)
{
    FILE *infile;
    char *opstr;
    char *fname = 0;
    int i;
    int size;
    //int parms[4];
    int loopcount = 0;
    int maxloops = 2000000000;
    int runflag;

    for (i = 1; i < argc; i++)
    {
	if (strcmp(argv[i], "-l") == 0)
	    printflag = 1;
	else if (strcmp(argv[i], "-s") == 0)
	    symflag = 1;
	else if (argv[i][0] == '-' && argv[i][1] >= '0' && argv[i][1] <= '9')
	    sscanf(argv[i] + 1, "%d", &maxloops);
	else if (argv[i][0] == '-')
	    usage();
	else if (!fname)
	    fname = argv[i];
	else
	    usage();
    }

    bootfile = fname;

    if (!fname) usage();

    ReadOpsFile();

    reboot();

    runflag = 1;
    while (runflag && ++loopcount < maxloops)
    {
	runflag = 0;
	for (i = 0; i < 8; i++)
	{
	    if (CogVars[i].state)
	    {
	        if (LONG(0xf00c) && CogVars[i].state == 1)
		{
		    long dcurr = CogVars[i].dcurr;
		    printf("Cog %d: %4.4x %8.8x - ", i, dcurr, LONG(dcurr - 4));
		    PrintOp(&CogVars[i]);
		}
	        ExecuteOp(&CogVars[i]);
		runflag = 1;
	    }
            CheckCommand();
	}
    }
}
